==> CRTL-0.13.4-py3-none-any.whl: crtl/s3_file_utils/s3_file_utils.py <==
"""
Utilities for working with the local dataset cache.
Copied from https://github.com/trend-community/krystie/blob/master/fileutils/fileutils.py
"""

import gzip
import io
import os
import pathlib
import shutil
import tempfile
import json
from urllib.parse import urlparse
from pathlib import Path
from typing import Optional, Tuple, Union, IO
from hashlib import sha256
import boto3
from spacy.util import registry

CACHE_ROOT = Path(os.getenv("CRTL_CACHE", Path.home()/".crtl"))

s3 = boto3.client("s3")
class S3Url(object):
    def __init__(self, url):
        self._parsed = urlparse(url, allow_fragments=False)
        print(self.bucket)
        print(self.key)

    @property
    def bucket(self):
        return self._parsed.netloc

    @property
    def key(self):
        if self._parsed.query:
            return self._parsed.path.lstrip("/") + "?" + self._parsed.query
        else:
            return self._parsed.path.lstrip("/")

    @property
    def url(self):
        return self._parsed.geturl()

    @property
    def head_object(self):
        return s3.head_object(Bucket=self.bucket, Key=self.key)

@registry.readers("cached_path")
def cached_path(url_or_filepath: Union[str,Path], cache_dir: Union[str,Path] = CACHE_ROOT) -> Path:
    """
    Given something that might be a URL (or might be a local path),
    determine which. If it's a URL, download the file and cache it, and
    return the path to the cached file. If it's already a local path,
    make sure the file exists and then return the path.
    """
    if isinstance(url_or_filepath,Path):
        if not url_or_filepath.exists():
            raise FileNotFoundError("file {} not found".format(url_or_filepath))
        return url_or_filepath
    else:
        parsed = urlparse(str(url_or_filepath))
        if parsed.scheme == "s3":
            # S3URL, so get it from the cache (downloading if necessary)
            return get_from_cache(S3Url(url_or_filepath), cache_dir)
        elif parsed.scheme == "":
            #File 
            file_path =Path(url_or_filepath)
            if file_path.exists():
                return file_path
            else:
                raise FileNotFoundError("file {} not found".format(url_or_filepath))
        else:
            # Something unknown
            raise ValueError(
                "unable to parse {} as a URL or as a local path".format(url_or_filepath)
            )



def s3_url_to_filename(url:str, etag: Optional[str] = None) -> Path:
    """
    Convert `url` into a hashed filename in a repeatable way.
    If `etag` is specified, append its hash to the url's, delimited
    by a period.
    """

    last_part = url.split("/")[-1]
    url_bytes = url.encode("utf-8")
    url_hash = sha256(url_bytes)
    filename = url_hash.hexdigest()

    if etag:
        etag_bytes = etag.encode("utf-8")
        etag_hash = sha256(etag_bytes)
        filename += "." + etag_hash.hexdigest()

    filename += "." + last_part
    return Path(filename)


def filename_to_url(filename: str, cache_dir: Path = CACHE_ROOT) -> Tuple[str, Optional[str]]:
    """
    Return the url and etag (which may be ``None``) stored for `filename`.
    Raise ``FileNotFoundError`` if `filename` or its stored metadata do not exist.
    """

    cache_path = cache_dir / filename
    
    if not cache_path.exists:
        raise FileNotFoundError("file {} not found".format(cache_path))

    meta_path = cache_dir.with_suffix(".json")
    if not meta_path.exists:
        raise FileNotFoundError("file {} not found".format(meta_path))

    with open(meta_path) as meta_file:
        metadata = json.load(meta_file)
    url = metadata["url"]
    etag = metadata["etag"]

    return url, etag


def boto3_get(url: S3Url, temp_file: IO) -> None:
    req = s3.get_object(Bucket=url.bucket, Key=url.key)["Body"]
    for chunk in req.iter_chunks(chunk_size=1024):
        if chunk:  # filter out keep-alive new chunks
            temp_file.write(chunk)


def get_from_cache(s3_url: S3Url, cache_dir: Union[Path,str] = CACHE_ROOT) -> Path:
    """
    Given a URL str, look for the corresponding dataset in the local cache.
    If it's not there, download it. Then return the path to the cached file.
    """
    cache_dir_path:Path = Path(cache_dir)

    cache_dir_path.mkdir(exist_ok=True)

    header = s3_url.head_object
    print(header)
    etag = header["ResponseMetadata"]["HTTPHeaders"]["etag"]
    status_code = header["ResponseMetadata"]["HTTPStatusCode"]
    if status_code != 200:
        raise IOError(
            "HEAD request failed for url {} with status code {}".format(
                s3_url.url, status_code
            )
        )
    filename = s3_url_to_filename(s3_url.url, etag)
    # get cache path to put the file
    cache_file_path = cache_dir_path / filename 
    if not cache_file_path.exists():
        # Download to temporary file, then copy to cache dir once finished.
        # Otherwise you get corrupt cache entries if the download gets interrupted.
        # type: IO
        download_to_cache(s3_url, cache_file_path, etag)

    return cache_file_path

def download_to_cache(s3_url:S3Url, cache_file_path:Path, etag:str):
    with tempfile.NamedTemporaryFile() as temp_file: 
        print(f"{s3_url.url} not found in cache, downloading to {temp_file.name}")
            # GET file object
        boto3_get(s3_url, temp_file)

            # we are copying the file before closing it, so flush to avoid truncation
        temp_file.flush()
            # shutil.copyfileobj() starts at the current position, so go to the start
        temp_file.seek(0)

        print(
                f"Finished download, copying {temp_file.name} to cache at {cache_file_path}"
            )
        with open(cache_file_path, "wb") as cache_file:
            shutil.copyfileobj(temp_file, cache_file)

        meta = {"url": s3_url.url, "etag": etag}
        meta_path = cache_file_path.with_suffix(".json")
        with open(meta_path, "w") as meta_file:
            json.dump(meta, meta_file)

def upload_gzipped(bucket_str:str, key:str, fp:pathlib.Path, compressed_fp=None, content_type='text/plain'):
    """
    Compress and upload the contents from fp to S3.
    If compressed_fp is None, the compression is performed in memory.
    """
    if not compressed_fp:
        compressed_fp = io.BytesIO()
    with open(fp,"rb") as f:
        with gzip.GzipFile(fileobj=compressed_fp, mode='wb') as gz:
            shutil.copyfileobj(f, gz)
    compressed_fp.seek(0)
    s3.upload_fileobj(
        compressed_fp,
        bucket_str,
        key,
        ExtraArgs={'ContentType': content_type, 'ContentEncoding': 'gzip'})
        
# if __name__ == "__main__":
#     read_gzipped_jsonl(cached_path("https://databrickstrenddata/lkb/crtl/trendnet.gzip"))

==> CRTL-0.16.3-py3-none-any.whl: crtl/s3_file_utils/s3_file_utils.py <==
"""
Utilities for working with the local dataset cache.
Copied from https://github.com/trend-community/krystie/blob/master/fileutils/fileutils.py
"""

import gzip
import io
import os
import pathlib
import shutil
import tempfile
import json
from urllib.parse import urlparse
from pathlib import Path
from typing import Optional, Tuple, Union, IO
from hashlib import sha256
import boto3
from spacy.util import registry

CACHE_ROOT = Path(os.getenv("CRTL_CACHE", Path.home() / ".crtl"))

s3 = boto3.client("s3")


class S3Url(object):
    def __init__(self, url):
        self._parsed = urlparse(url, allow_fragments=False)
        print(self.bucket)
        print(self.key)

    @property
    def bucket(self):
        return self._parsed.netloc

    @property
    def key(self):
        if self._parsed.query:
            return self._parsed.path.lstrip("/") + "?" + self._parsed.query
        else:
            return self._parsed.path.lstrip("/")

    @property
    def url(self):
        return self._parsed.geturl()

    @property
    def head_object(self):
        return s3.head_object(Bucket=self.bucket, Key=self.key)


@registry.readers("cached_path")
def cached_path(
    url_or_filepath: Union[str, Path], cache_dir: Union[str, Path] = CACHE_ROOT
) -> Path:
    """
    Given something that might be a URL (or might be a local path),
    determine which. If it's a URL, download the file and cache it, and
    return the path to the cached file. If it's already a local path,
    make sure the file exists and then return the path.
    """
    if isinstance(url_or_filepath, Path):
        if not url_or_filepath.exists():
            raise FileNotFoundError("file {} not found".format(url_or_filepath))
        return url_or_filepath
    else:
        parsed = urlparse(str(url_or_filepath))
        if parsed.scheme == "s3":
            # S3URL, so get it from the cache (downloading if necessary)
            return get_from_cache(S3Url(url_or_filepath), cache_dir)
        elif parsed.scheme == "":
            # File
            file_path = Path(url_or_filepath)
            if file_path.exists():
                return file_path
            else:
                raise FileNotFoundError("file {} not found".format(url_or_filepath))
        else:
            # Something unknown
            raise ValueError(
                "unable to parse {} as a URL or as a local path".format(url_or_filepath)
            )


def s3_url_to_filename(url: str, etag: Optional[str] = None) -> Path:
    """
    Convert `url` into a hashed filename in a repeatable way.
    If `etag` is specified, append its hash to the url's, delimited
    by a period.
    """

    last_part = url.split("/")[-1]
    url_bytes = url.encode("utf-8")
    url_hash = sha256(url_bytes)
    filename = url_hash.hexdigest()

    if etag:
        etag_bytes = etag.encode("utf-8")
        etag_hash = sha256(etag_bytes)
        filename += "." + etag_hash.hexdigest()

    filename += "." + last_part
    return Path(filename)


def filename_to_url(
    filename: str, cache_dir: Path = CACHE_ROOT
) -> Tuple[str, Optional[str]]:
    """
    Return the url and etag (which may be ``None``) stored for `filename`.
    Raise ``FileNotFoundError`` if `filename` or its stored metadata do not exist.
    """

    cache_path = cache_dir / filename

    if not cache_path.exists:
        raise FileNotFoundError("file {} not found".format(cache_path))

    meta_path = cache_dir.with_suffix(".json")
    if not meta_path.exists:
        raise FileNotFoundError("file {} not found".format(meta_path))

    with open(meta_path) as meta_file:
        metadata = json.load(meta_file)
    url = metadata["url"]
    etag = metadata["etag"]

    return url, etag


def boto3_get(url: S3Url, temp_file: IO) -> None:
    req = s3.get_object(Bucket=url.bucket, Key=url.key)["Body"]
    for chunk in req.iter_chunks(chunk_size=1024):
        if chunk:  # filter out keep-alive new chunks
            temp_file.write(chunk)


def get_from_cache(s3_url: S3Url, cache_dir: Union[Path, str] = CACHE_ROOT) -> Path:
    """
    Given a URL str, look for the corresponding dataset in the local cache.
    If it's not there, download it. Then return the path to the cached file.
    """
    cache_dir_path: Path = Path(cache_dir)

    cache_dir_path.mkdir(exist_ok=True)

    header = s3_url.head_object
    print(header)
    etag = header["ResponseMetadata"]["HTTPHeaders"]["etag"]
    status_code = header["ResponseMetadata"]["HTTPStatusCode"]
    if status_code != 200:
        raise IOError(
            "HEAD request failed for url {} with status code {}".format(
                s3_url.url, status_code
            )
        )
    filename = s3_url_to_filename(s3_url.url, etag)
    # get cache path to put the file
    cache_file_path = cache_dir_path / filename
    if not cache_file_path.exists():
        # Download to temporary file, then copy to cache dir once finished.
        # Otherwise you get corrupt cache entries if the download gets interrupted.
        # type: IO
        download_to_cache(s3_url, cache_file_path, etag)

    return cache_file_path


def download_to_cache(s3_url: S3Url, cache_file_path: Path, etag: str):
    with tempfile.NamedTemporaryFile() as temp_file:
        print(f"{s3_url.url} not found in cache, downloading to {temp_file.name}")
        # GET file object
        boto3_get(s3_url, temp_file)

        # we are copying the file before closing it, so flush to avoid truncation
        temp_file.flush()
        # shutil.copyfileobj() starts at the current position, so go to the start
        temp_file.seek(0)

        print(
            f"Finished download, copying {temp_file.name} to cache at {cache_file_path}"
        )
        with open(cache_file_path, "wb") as cache_file:
            shutil.copyfileobj(temp_file, cache_file)

        meta = {"url": s3_url.url, "etag": etag}
        meta_path = cache_file_path.with_suffix(".json")
        with open(meta_path, "w") as meta_file:
            json.dump(meta, meta_file)


def upload_gzipped(
    bucket_str: str,
    key: str,
    fp: pathlib.Path,
    compressed_fp=None,
    content_type="text/plain",
):
    """
    Compress and upload the contents from fp to S3.
    If compressed_fp is None, the compression is performed in memory.
    """
    if not compressed_fp:
        compressed_fp = io.BytesIO()
    with open(fp, "rb") as f:
        with gzip.GzipFile(fileobj=compressed_fp, mode="wb") as gz:
            shutil.copyfileobj(f, gz)
    compressed_fp.seek(0)
    s3.upload_fileobj(
        compressed_fp,
        bucket_str,
        key,
        ExtraArgs={"ContentType": content_type, "ContentEncoding": "gzip"},
    )


# if __name__ == "__main__":
#     read_gzipped_jsonl(cached_path("https://databrickstrenddata/lkb/crtl/trendnet.gzip"))
